"""
A Custom CloudFormation provider supporting the latest version of
AWS::ElasticLoadBalancingV2::ListenerRule.
"""
import boto3

import cfnresponse

ELB = boto3.client('elbv2')


def lambda_handler(event, context):
    """
    Implements the Custom CloudFormation Resource interface
    """
    arn = None
    try:
        method, arn, kwargs = get_arguments(event)
        if method == 'create_rule':
            response = ELB.create_rule(**kwargs)
            arn = response['Rules'][0]['RuleArn']
        elif method == 'modify_rule':
            ELB.modify_rule(**kwargs)
        elif method == 'delete_rule':
            try:
                ELB.delete_rule(**kwargs)
            except Exception as error:
                print('{}'.format(error))

        return cfnresponse.send(event, context, cfnresponse.SUCCESS, {}, arn)
    except Exception as error:
        print('{}'.format(error))
        return cfnresponse.send(event, context, cfnresponse.FAILED, {}, arn if arn else 'failed-to-create')


def get_arguments(event):
    """
    retrieve method, ARN and resource arguments appropriate for the API
    :param event: containing the CFN request
    :return:  method, arn, arguments - for the api call of the rule
    """
    request_type = event['RequestType']
    arn = event['PhysicalResourceId'] if 'PhysicalResourceId' in event else None

    if request_type == 'Delete':
        return 'delete_rule', arn, {'RuleArn': arn}
    else:
        arguments = event['ResourceProperties'].copy()
        convert_from_string(arguments, ["Order", "Priority", "SessionTimeout"])

        if 'ServiceToken' in arguments:
            del arguments['ServiceToken']

        if is_create_required(event):
            return 'create_rule', None, arguments
        else:
            arguments['RuleArn'] = arn
            if 'ListenerArn' in arguments:
                del arguments['ListenerArn']
            if 'Priority' in arguments:
                del arguments['Priority']

        return 'modify_rule', arn, arguments


def convert_from_string(properties, fields):
    """
    convert the specified fields to native type.
    """
    for name in properties:
        if isinstance(properties[name], dict):
            convert_from_string(properties[name], fields)
        elif name in fields and isinstance(properties[name], str):
            value = str(properties[name])
            if value == 'true':
                properties[name] = True
            elif value == 'false':
                properties[name] = False
            elif value.isdigit():
                properties[name] = int(value)
            else:
                pass  # leave it a string.


def is_create_required(event):
    """
    Indicates whether a create is required in order to update.
    :param event: to check
    """
    if event['RequestType'] == 'Create':
        return True

    new = event['ResourceProperties']
    old = event['OldResourceProperties'] if 'OldResourceProperties' in event else new

    return old.get('Priority', 0) != new.get('Priority', 0) or old.get('ListenerArn', None) != new.get('ListenerArn',
                                                                                                       None)



import boto3
import botocore

from aws_cfn_update.packer_ami_updater import PackerAMIUpdater
from unittest import TestCase


class TestAMIUpdater(TestCase):
    def setUp(self):
        self.session = botocore.session.get_session(
            {"region": ("region", "AWS_DEFAULT_REGION", "eu-central-1", None)}
        )
        self.session.set_credentials(self, "deadbeef", "deadbeef")
        boto3.setup_default_session(botocore_session=self.session)
        assert boto3.DEFAULT_SESSION

    def tearDown(self) -> None:
        boto3.DEFAULT_SESSION = None

    def test_is_source_filter_name_match(self):
        updater = PackerAMIUpdater()

        source_ami_filter = {"filters": {"name": "Windows_Server 2016 2020.02.10"}}
        updater.ami_name_pattern = "Windows_Server 2016*"
        assert updater.is_source_filter_name_match(source_ami_filter)

        source_ami_filter = {"filters": {"blabla": "Windows_Server 2016 2020.02.10"}}
        assert not updater.is_source_filter_name_match(source_ami_filter)
        assert not updater.is_source_filter_name_match({})

    def test_create_describe_image_request(self):
        source_ami_filter = {
            "filters": {
                "virtualization-type": "hvm",
                "name": "Windows_Server-2016-English-Full-Base-2020.01.10",
                "root-device-type": "ebs",
            },
            "owners": ["801119661308"],
        }
        updater = PackerAMIUpdater()
        updater.ami_name_pattern = "Windows_Server-2016-English-Full-Base-*"
        response = updater.create_describe_image_request(source_ami_filter)
        assert response["Owners"] == source_ami_filter["owners"]
        assert has_filter(response["Filters"], "virtualization-type", ["hvm"])
        assert has_filter(response["Filters"], "name", [updater.ami_name_pattern])
        assert has_filter(response["Filters"], "state", ["available"])
        assert has_filter(response["Filters"], "root-device-type", ["ebs"])

        source_ami_filter = {
            "filters": {
                "virtualization-type": "hvm",
                "name": "Windows_Server-2016-English-Full-Base-2020.01.10",
                "root-device-type": "ebs",
            }
        }
        response = updater.create_describe_image_request(source_ami_filter)
        assert not response.get("Owners")


def has_filter(filters, name, values):
    return next(
        filter(lambda f: f["Name"] == name and f["Values"] == values, filters), None
    )



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import re
import sys
import os.path
import json
import collections

from ruamel.yaml import YAML


class CfnUpdater(object):
    """
    base class for a CloudFormation  update. To implement a specific updater:

    - inherit this class
    - override the method update_template()
    - call the method update(path)

    it will read the template from files with extension .yaml, .yml and .json into
    the property `self.template`.

    If the property `self.dirty` is set to True, the template will be written
    back to the originating file.

    Please note that formatting and comments may be lost, when using this
    updater.
    """

    def __init__(self):
        self.basename = None
        self.template = False
        self.template = None
        self.template_format = None
        self.dirty = False
        self.dry_run = False
        self.verbose = False
        self._filename = None
        self.yaml = YAML(typ="rt")
        self.yaml.preserve_quotes = True
        self.yaml.explicit_start = True
        self.yaml.width = 4096
        self.yaml.indent(mapping=2, sequence=4, offset=2)

    @property
    def filename(self):
        """
        current template filename
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        """
        requires `filename` ends with `.json`, `.yaml` or `.yml`.
        sets `basename`, `template_format` and `_filename` accordingly.
        clears `dirty` and `templae`
        """
        self._filename = filename
        parts = os.path.splitext(os.path.basename(filename))
        self.basename = parts[0]
        self.template_format = parts[1]
        self.template = None
        self.dirty = False
        if self.template_format not in (".json", ".yml", ".yaml"):
            raise ValueError("%s has no .json, .yaml or .yml extension." % filename)

    def load(self):
        """
        loads `template` from `filename` as dictionary.
        """
        self.dirty = False
        self.template = None
        with open(self.filename, "r") as f:
            if self.template_format == ".json":
                self.template = json.load(f, object_pairs_hook=collections.OrderedDict)
            else:
                self.template = self.yaml.load(f)

    def is_cloudformation_template(self):
        """
        returns true if the `self.template` is a AWS CloudFormation template
        """
        return self.template and "AWSTemplateFormatVersion" in self.template

    def write(self):
        """
        write modified content from `template` to `filename`. It will retain it's original
        format (yaml or json) but loose original formatting and comments.
        """
        if not self.dirty:
            if self.verbose:
                sys.stderr.write("INFO: no changes in {}\n".format(self.filename))
            return

        if self.dry_run:
            return

        with open(self.filename, "w") as f:
            if self.template_format == ".yaml":
                self.yaml.dump(self.template, f)
            else:
                json.dump(self.template, f, separators=(",", ": "), indent=2)

    def update_template(self):
        """
        implement the update logic of `self.template`. Set self.dirty to True, if you modified.
        """
        pass

    @property
    def resources(self):
        return self.template.get("Resources", {})

    def update(self, path):
        """
        recursively updates all the cloudformation templates in the specified `path`. `path` may be a file,
        a directory or a list of paths.
        """
        if isinstance(path, (list, tuple)):
            for p in path:
                self.update(p)
        elif os.path.isfile(path):
            if (
                path.endswith(".yml")
                or path.endswith(".yaml")
                or path.endswith(".json")
            ):
                self.filename = path
                self.load()
                if self.is_cloudformation_template():
                    self.update_template()
                    self.write()
                else:
                    if self.verbose:
                        sys.stderr.write(
                            "INFO: skipping {} as it is not a CloudFormation template\n".format(
                                path
                            )
                        )
        elif os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for f in files:
                    self.update(os.path.join(root, f))
        else:
            sys.stderr.write("ERROR: {} is not a file or directory\n".format(path))
            sys.exit(1)


def read_template(filename: str) -> dict:
    src = CfnUpdater()
    src.filename = filename
    src.load()
    return src.template



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2024 binx.io B.V.
import re
import sys
from typing import Optional

from .cfn_updater import CfnUpdater

_s3_key_semver_pattern = re.compile(
    r"(.*)(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?\.zip$"
)


class LambdaS3KeyUpdater(CfnUpdater):
    """
        Updates the S3Key entry of a Lambda Function definition. The s3 key should
        match the pattern <prefix><semver>.zip.

    \b
      ELBListenerRuleProvider:
        Type: AWS::Lambda::Function
        Properties:
          Code:
            S3Bucket: !Sub 'binxio-public-${AWS::Region}'
            S3Key: lambdas/iam-sudo-0.1.0.zip
    \b
      ELBListenerRuleProvider:
        Type: AWS::Lambda::Function
        Properties:
          Code:
            S3Bucket: !Sub 'binxio-public-${AWS::Region}'
            S3Key: lambdas/iam-sudo-0.3.1.zip
              ...

        The environment variable AWS_CFN_UPDATE_LAMBDA_S3_KEYS can be used to specify a
        whitespace separated list of S3 keys to update.
    """

    def __init__(self):
        super().__init__()
        self._s3_keys = {}

    @property
    def s3_keys(self) -> [str]:
        return self._s3_keys.values()

    def s3_key_by_prefix(self, prefix: str) -> Optional[str]:
        return self._s3_keys.get(prefix)

    @s3_keys.setter
    def s3_keys(self, s3_keys: [str]):
        self._s3_keys = {}
        for s3_key in s3_keys:
            match = _s3_key_semver_pattern.match(s3_key)
            if not match:
                raise ValueError(f"{s3_key} is not a semver S3Key")
            self._s3_keys[match.group(1)] = s3_key

    def update_template(self):
        """
        updates the S3Key property of a AWS::Lambda::Function definition.

        """
        for resource_name, resource in self.template.get("Resources", {}).items():
            if resource and resource["Type"] == "AWS::Lambda::Function":
                code = resource.get("Properties", {}).get("Code", {})
                s3_key = code.get("S3Key")
                if not s3_key:
                    return

                match = _s3_key_semver_pattern.match(s3_key)
                if match and (replacement := self.s3_key_by_prefix(match.group(1))):
                    if replacement != s3_key:
                        sys.stderr.write(
                            "INFO: updating S3Key of Lambda function {} in {}\n".format(
                                resource_name, self.filename
                            )
                        )

                        code["S3Key"] = replacement
                        self.dirty = True

    def main(self, s3_keys: [str], paths, dry_run, verbose):
        self.s3_keys = s3_keys
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from typing import Optional, Union, List

from .cfn_updater import CfnUpdater


class ContainerImageUpdater(CfnUpdater):
    """
        Updates the Docker image of ECS Container Definitions.

        it will update any container definition where the base image name matches
        the specified image name excluding the tag.

        For example, an image name of `mvanholsteijn/paas-monitor:0.6.0` will update:

    \b
            Type: AWS::ECS::TaskDefinition
            Properties:
              ContainerDefinitions:
                - Name: paas-monitor
                  Image: mvanholsteijn/paas-monitor:0.5.9

        to:

    \b
            Type: AWS::ECS::TaskDefinition
            Properties:
              ContainerDefinitions:
                - Name: paas-monitor
                  Image: mvanholsteijn/paas-monitor:0.6.0


        The environment variable AWS_CFN_UPDATE_CONTAINER_IMAGES can be used to specify a
        whitespace separated list of container images to update.
    """

    def __init__(self):
        super(ContainerImageUpdater, self).__init__()
        self._images = {}

    @property
    def images(self) -> [str]:
        return self._images.values()

    @images.setter
    def images(self, images: [str]):
        self._images = {}
        for image in images:
            parts = image.split(":")
            if len(parts) != 2:
                raise ValueError(f"{image} is an invalid image name")
            base = parts[0]
            if base in self._images and image != self._images[base]:
                raise ValueError(f"image already defined for {base}")
            self._images[base] = image

    def get_new_image_reference(self, image: str) -> Optional[str]:
        return self._images.get(image.split(":")[0])

    @staticmethod
    def is_task_definition(resource):
        """
        returns true if the resource is of type AWS::ECS::TaskDefinition
        """
        return resource.get("Type", "") == "AWS::ECS::TaskDefinition"

    def all_matching_task_definitions(self, resources):
        return filter(lambda n: self.is_task_definition(resources[n]), resources)

    def update_template(self):
        """
        updates the Image of container definitions in the CFN template `self.template`.
        """
        resources = self.template.get("Resources", {})
        for task_name in self.all_matching_task_definitions(resources):
            task = resources[task_name]
            containers = task.get("Properties", {}).get("ContainerDefinitions", [])
            for container in filter(
                lambda c: self.get_new_image_reference(c["Image"]),
                filter(lambda c: isinstance(c.get("Image", None), str), containers),
            ):
                new_image = self.get_new_image_reference(container["Image"])
                if container["Image"] != new_image:
                    sys.stderr.write(
                        "INFO: updating image of container definition {} for task {} in {}\n".format(
                            container["Name"], task_name, self.filename
                        )
                    )
                    container["Image"] = new_image
                    self.dirty = True

    def main(self, image: [str], dry_run: bool, verbose: bool, paths: [str]):
        self.images = image
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



import os
import json
from aws_cfn_update.latest_ami_updater import AMIUpdater, make_new_resource_name


def test_new_resource_name():
    updater = stubbed_ami_updater()
    result = make_new_resource_name("WhatEverBaseAMI")
    assert result == "WhatEverBaseAMIv1"

    result = make_new_resource_name(result)
    assert result == "WhatEverBaseAMIv2"

    result = make_new_resource_name("WhatEv1rBaseAMIv3")
    assert result == "WhatEv1rBaseAMIv4"


def test_ami_update_inplace():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.a-amazon-ecs-optimized"}
                },
            }
        }
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.update_template()
    assert updater.dirty
    assert template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )


def test_with_other_filters():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {
                        "name": "amzn-ami-2013.09.a-amazon-ecs-optimized",
                        "owner-alias": "amazon",
                        "is-public": "true",
                    },
                    "Owners": ["amazon"],
                },
            }
        }
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.update_template()
    assert updater.dirty
    assert template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )


def test_no_matching_ami_found():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {
                        "name": "amzn-ami-2013.09.a-amazon-ecs-optimized",
                        "owner-alias": "amazon",
                        "is-public": "false",
                    },
                    "Owners": ["microsoft"],
                },
            }
        }
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.update_template()
    assert not updater.dirty
    assert template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )


def test_add_new_version():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.a-amazon-ecs-optimized"}
                },
            }
        },
        "Outputs": {"AMI": {"Ref": "CustomAMI"}},
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.add_new_version = True
    updater.update_template()
    assert updater.dirty
    assert "CustomAMIv1" in template["Resources"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomAMIv1"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert template["Outputs"]["AMI"]["Ref"] == "CustomAMIv1"


def test_add_new_version_partitions():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.a-amazon-ecs-optimized"}
                },
            },
            "CustomBMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2016.09.a-amazon-ecs-optimized"}
                },
            },
        },
        "Outputs": {"AMI": {"Ref": "CustomAMI"}, "BMI": {"Ref": "CustomBMI"}},
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.add_new_version = True
    updater.update_template()
    assert updater.dirty
    assert "CustomAMIv1" in template["Resources"]
    assert "CustomBMIv1" in template["Resources"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomBMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2016.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomAMIv1"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomBMIv1"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2016.09.a-amazon-ecs-optimized"
    )
    assert template["Outputs"]["AMI"]["Ref"] == "CustomAMIv1"
    assert template["Outputs"]["BMI"]["Ref"] == "CustomBMIv1"


def test_add_new_version_partitions_multiple_versions():
    template = {
        "Resources": {
            "CustomAMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.a-amazon-ecs-optimized"}
                },
            },
            "CustomAMIv2": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.b-amazon-ecs-optimized"}
                },
            },
            "CustomBMI": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2016.09.a-amazon-ecs-optimized"}
                },
            },
        },
        "Outputs": {"AMI": {"Ref": "CustomAMI"}, "BMI": {"Ref": "CustomBMI"}},
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.add_new_version = True
    updater.update_template()
    assert updater.dirty
    assert "CustomAMIv3" in template["Resources"]
    assert "CustomBMIv1" in template["Resources"]
    assert (
        template["Resources"]["CustomAMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomBMI"]["Properties"]["Filters"]["name"]
        == "amzn-ami-2016.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomAMIv3"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2013.09.a-amazon-ecs-optimized"
    )
    assert (
        template["Resources"]["CustomBMIv1"]["Properties"]["Filters"]["name"]
        > "amzn-ami-2016.09.a-amazon-ecs-optimized"
    )
    assert template["Outputs"]["AMI"]["Ref"] == "CustomAMIv3"
    assert template["Outputs"]["BMI"]["Ref"] == "CustomBMIv1"


def test_add_new_versions_keep_old_refs():
    template = {
        "Resources": {
            "CustomAMIv2": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.b-amazon-ecs-optimized"}
                },
            },
            "CustomAMIv1": {
                "Type": "Custom::AMI",
                "Properties": {
                    "Filters": {"name": "amzn-ami-2013.09.a-amazon-ecs-optimized"}
                },
            },
        },
        "Outputs": {"OldAMI": {"Ref": "CustomAMIv1"}, "NewAMI": {"Ref": "CustomAMIv2"}},
    }
    updater = stubbed_ami_updater()
    updater.template = template
    updater.ami_name_pattern = "amzn-ami-*ecs-optimized"
    updater.add_new_version = True
    updater.update_template()
    assert updater.dirty
    assert "CustomAMIv3" in template["Resources"]
    assert template["Outputs"]["OldAMI"]["Ref"] == "CustomAMIv1"
    assert template["Outputs"]["NewAMI"]["Ref"] == "CustomAMIv3"


all_responses = []
with open(os.path.join(os.path.dirname(__file__), "dummy_responses.json"), "r") as f:
    all_dummy_responses = json.load(f)


def stubbed_ami_updater():
    result = AMIUpdater()
    result._describe_images = describe_images_stub
    return result


def describe_images_stub(**request):
    result = next(
        filter(lambda rr: rr["request"] == request, all_dummy_responses), None
    )
    if not result:
        assert False, "no dummy response found for this request"
    return result["response"]



from aws_cfn_update.config_rule_inline_code_updater import (
    ConfigRuleInlineCodeUpdater as Updater,
)
import json


sample = {
    "Resources": {
        "ConfigRule": {
            "Type": "AWS::Config::ConfigRule",
            "Properties": {
                "Source": {
                    "Owner": "CUSTOM_POLICY",
                    "CustomPolicyDetails": {
                        "EnableDebugLogDelivery": "true",
                        "PolicyRuntime": "guard-2.x.x",
                    },
                }
            },
        }
    }
}


def test_add_body():
    updater = Updater()
    updater.resource_name = "ConfigRule"
    updater.code = 'let buckets = Resources.*[ Type == "AWS::S3::Bucket" ]'
    updater.template = sample.copy()

    updater.update_template()
    assert updater.dirty
    assert "Properties" in updater.template["Resources"]["ConfigRule"]

    assert "Source" in updater.template["Resources"]["ConfigRule"]["Properties"]
    source = updater.template["Resources"]["ConfigRule"]["Properties"]["Source"]
    assert "CustomPolicyDetails" in source

    assert "CUSTOM_POLICY" == source["Owner"]
    assert "true" == source["CustomPolicyDetails"]["EnableDebugLogDelivery"]
    assert "guard-2.x.x" in source["CustomPolicyDetails"]["PolicyRuntime"]

    assert (
        source["CustomPolicyDetails"]["PolicyText"]
        == 'let buckets = Resources.*[ Type == "AWS::S3::Bucket" ]'
    )


def test_replace_body():
    updater = Updater()
    updater.resource_name = "ConfigRule"
    updater.code = 'let buckets = Resources.*[ Type == "AWS::S3::Bucket" ]'
    updater.template = sample.copy()
    updater.template["Resources"]["ConfigRule"]["Properties"]["Source"][
        "CustomPolicyDetails"
    ]["PolicyText"] = "Existing Policy"

    updater.update_template()
    assert updater.dirty
    assert "Properties" in updater.template["Resources"]["ConfigRule"]

    assert "Source" in updater.template["Resources"]["ConfigRule"]["Properties"]
    source = updater.template["Resources"]["ConfigRule"]["Properties"]["Source"]
    assert "CustomPolicyDetails" in source

    assert "CUSTOM_POLICY" == source["Owner"]
    assert "true" == source["CustomPolicyDetails"]["EnableDebugLogDelivery"]
    assert "guard-2.x.x" in source["CustomPolicyDetails"]["PolicyRuntime"]

    assert (
        source["CustomPolicyDetails"]["PolicyText"]
        == 'let buckets = Resources.*[ Type == "AWS::S3::Bucket" ]'
    )


skip_body_sample = {"Resources": {"ConfigRule": {"Type": "AWS::EC2::Instance"}}}


def test_skip_body():
    updater = Updater()
    updater.resource_name = "ConfigRule"
    updater.code = 'print("hello world!")'
    updater.template = skip_body_sample.copy()

    updater.update_template()
    assert not updater.dirty
    assert "Properties" not in updater.template["Resources"]["ConfigRule"]



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import copy
import fnmatch
import re
from collections import OrderedDict

import boto3
import sys

from .cfn_updater import CfnUpdater
from .replace_references import replace_references


def split_resource_name(resource_name):
    m = re.match(r"^(?P<base>.*)v(?P<version>[0-9]+)$", resource_name)

    if m is not None:
        base = m.group("base")
        version = int(m.group("version"))
    else:
        base = resource_name
        version = 0

    return base, version


def make_new_resource_name(resource_name):
    base, version = split_resource_name(resource_name)
    return "{}v{}".format(base, version + 1)


class AMIUpdater(CfnUpdater):
    """
        Updates the AMI name of Custom::AMI resources to the latest version.
        With an ami-name-pattern of `amzn-ami-*ecs-optimized` it will update
        the following resource definition from:

    \b
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.a-amazon-ecs-optimized
               Owners:
                 - amazon

        to:

    \b
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.l-amazon-ecs-optimized
               Owners:
                 - amazon


        By specifying --add-new-version, a new Custom::AMI will be added
        to the template with a new name. A suffix `v<version>` is appended
        to create the new resource. Any reference to the highest Custom::AMI version
        resource is replaced. It will change:

    \b
          CustomAMI:
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.a-amazon-ecs-optimized
               Owners:
                 - amazon
          CustomAMIv2:
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.b-amazon-ecs-optimized
               Owners:
                 - amazon
          Instance:
             Type: AWS::EC2::Instance
             Properties:
                ImageId: !Ref CustomAMI

        to:

    \b
          CustomAMI:
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.a-amazon-ecs-optimized
               Owners:
                 - amazon
          CustomAMIv2:
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.b-amazon-ecs-optimized
               Owners:
                 - amazon
          CustomAMIv3:
             Type: Custom::AMI
             Properties:
               Filters:
                 name: amzn-ami-2017.09.l-amazon-ecs-optimized
               Owners:
                 - amazon
          Instance:
             Type: AWS::EC2::Instance
             Properties:
                ImageId: !Ref CustomAMIv3
    """

    def __init__(self):
        super(AMIUpdater, self).__init__()
        self._ami_name_pattern = None
        self.add_new_version = False
        self.latest_ami_name_pattern = None

    @property
    def ami_name_pattern(self):
        return self._ami_name_pattern

    @ami_name_pattern.setter
    def ami_name_pattern(self, name):
        self._ami_name_pattern = name

    @staticmethod
    def is_ami_definition(resource):
        return resource.get("Type", "") == "Custom::AMI"

    def create_describe_image_request(self, ami):
        # copy the filters values, except for name and state.
        properties = ami.get("Properties", {})
        filters = [
            {"Name": "name", "Values": [self.ami_name_pattern]},
            {"Name": "state", "Values": ["available"]},
        ]
        for k, v in properties.get("Filters", {}).items():
            if k not in ["name", "state"]:
                filters.append({"Name": k, "Values": v if isinstance(v, list) else [v]})

        # copy the rest of the arguments
        result = {
            n: (
                properties.get(n)
                if isinstance(properties.get(n), list)
                else [properties.get(n)]
            )
            for n in filter(
                lambda n: properties.get(n) is not None,
                ["Owners", "ImageIds", "ExecutableUsers"],
            )
        }
        result["Filters"] = filters
        return result

    def _describe_images(self, **kwargs):
        return boto3.client("ec2").describe_images(**kwargs)

    def load_latest_ami_name_pattern(self, resource):
        response = self._describe_images(**self.create_describe_image_request(resource))

        images = sorted(response["Images"], key=lambda i: i["CreationDate"])
        self.latest_ami_name_pattern = images[-1]["Name"] if len(images) > 0 else None
        if len(images) > 0:
            sys.stderr.write(
                "INFO: using {} matching {}, created on {}\n".format(
                    self.latest_ami_name_pattern,
                    self.ami_name_pattern,
                    images[-1]["CreationDate"],
                )
            )

    def is_ami_name_pattern_match(self, name):
        regex = re.compile(fnmatch.translate(self.ami_name_pattern))
        return regex.match(name)

    def all_custom_ami_resources(self):
        r = self.resources
        return filter(
            lambda n: self.is_ami_definition(r[n]) and self.is_name_filter_match(r[n]),
            r,
        )

    def is_name_filter_match(self, ami):
        if (
            "Properties" in ami
            and "Filters" in ami["Properties"]
            and "name" in ami["Properties"]["Filters"]
        ):
            return self.is_ami_name_pattern_match(ami["Properties"]["Filters"]["name"])
        return False

    def custom_ami_resources_partitions(self):
        partitions = {}
        ami_resources = self.all_custom_ami_resources()
        for resource_name in ami_resources:
            base, _ = split_resource_name(resource_name)
            if base not in partitions:
                partitions[base] = {}
            partitions[base][resource_name] = self.resources[resource_name]

        for base in partitions:
            resource_names = sorted(
                partitions[base].keys(), key=lambda n: split_resource_name(n)[1]
            )
            yield OrderedDict({name: partitions[base][name] for name in resource_names})

    @staticmethod
    def latest_custom_ami_resource(ami_resources):
        result = None
        highest_version = -1
        for resource_name in ami_resources:
            _, version = split_resource_name(resource_name)
            if version > highest_version:
                result = resource_name
                highest_version = version

        return result

    def ami_requires_update(self, ami):
        name = ami["Properties"]["Filters"]["name"]
        return (
            self.latest_ami_name_pattern is not None
            and name != self.latest_ami_name_pattern
        )

    def update_ami(self, resource_name, ami):
        if self.ami_requires_update(ami):
            ami["Properties"]["Filters"]["name"] = self.latest_ami_name_pattern
            sys.stderr.write(
                'INFO: updating AMI definition "{}" name filter to {} in {}\n'.format(
                    resource_name, self.latest_ami_name_pattern, self.filename
                )
            )
            self.dirty = True
        else:
            if self.verbose:
                sys.stderr.write(
                    'INFO: AMI definition "{}" name already up to date in {}\n'.format(
                        resource_name, self.filename
                    )
                )

    def update_inplace(self):
        """
        updates the name filter of AMI resource definitions in `self.template`.
        """
        for resource_name in self.all_custom_ami_resources():
            ami = self.resources[resource_name]
            self.load_latest_ami_name_pattern(ami)
            self.update_ami(resource_name, ami)

    def add_new_ami_resource(self):
        """
        add a new version of the AMI resource definition in `self.template`.
        """
        for ami_resources in self.custom_ami_resources_partitions():
            resource_name = self.latest_custom_ami_resource(ami_resources)
            if resource_name is not None:
                ami = copy.deepcopy(self.resources[resource_name])
                self.load_latest_ami_name_pattern(ami)
                if self.ami_requires_update(ami):
                    new_resource_name = make_new_resource_name(resource_name)
                    self.resources[new_resource_name] = ami
                    self.update_ami(new_resource_name, ami)
                    for old_resource_name in reversed(ami_resources):
                        if replace_references(
                            self.template, old_resource_name, new_resource_name
                        ):
                            break

    def update_template(self):
        if self.add_new_version:
            self.add_new_ami_resource()
        else:
            self.update_inplace()

    def main(self, ami_name_pattern, dry_run, verbose, add_new_version, path):
        self.dry_run = dry_run
        self.verbose = verbose
        self.ami_name_pattern = ami_name_pattern
        self.add_new_version = add_new_version
        self.load_latest_ami_name_pattern({})
        if self.latest_ami_name_pattern is None:
            sys.stderr.write(
                "ERROR: image name {} does not resolve to an active AMI \n".format(
                    self.ami_name_pattern
                )
            )
            sys.exit(1)

        self.update(path)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import re
import sys
from datetime import datetime, timedelta, tzinfo

from croniter import croniter
from tzlocal import get_localzone

from .cfn_updater import CfnUpdater


class CronScheduleExpressionUpdater(CfnUpdater):
    """
        Updates the schedule expression of an AWS::Events::Rules resources to reflect the
        scheduled time in UTC. The required cron rule is taken from the description. It will
        update the following resource definition from:

    \b
          DailyTaskSchedule:
            Type: AWS::Events::Rule
            Properties:
              Description: run daily - cron(30 01 * * ? *)
              Name: run daily
              ScheduleExpression: cron(30 01 * * ? *)
              State: ENABLED

    to

    \b
          DailyTaskSchedule:
            Type: AWS::Events::Rule
            Properties:
              Description: run daily - cron(30 01 * * ? *)
              Name: run daily
              ScheduleExpression: cron(30 23 * * ? *)
              State: ENABLED

    with --timezone Europe/Amsterdam and --date 2018-08-01. If the updater is run with --date 2018-12-01, it
    will change it to:

    \b
          DailyTaskSchedule:
            Type: AWS::Events::Rule
            Properties:
              Description: run daily - cron(30 01 * * ? *)
              Name: run daily
              ScheduleExpression: cron(30 00 * * ? *)
              State: ENABLED


    """

    def __init__(self):
        super(CronScheduleExpressionUpdater, self).__init__()
        self._timezone = get_localzone()

        self._today = datetime.now()
        self.verbose = False
        self.dry_run = False

    @property
    def timezone(self):
        return self._timezone

    @timezone.setter
    def timezone(self, timezone):
        self._timezone = timezone
        self._today = timezone.localize(self._today)

    @property
    def today(self):
        return self._today

    @today.setter
    def today(self, dt):
        if dt.tzinfo:
            self._today = dt.replace(tzinfo=self.timezone)
        else:
            self._today = self.timezone.localize(dt)

    def is_matching_resource(self, resource):
        aws_type = resource.get("Type", "")
        description = resource.get("Properties", {}).get("Description", "")
        expression = resource.get("Properties", {}).get("ScheduleExpression", None)

        if aws_type == "AWS::Events::Rule" and aws_cron_pattern.search(description):
            return expression is None or aws_cron_pattern.search(expression)

        return False

    def all_matching_resources(self, resources):
        return {k: v for k, v in resources.items() if self.is_matching_resource(v)}

    def update_template(self):
        """
        converts the cron expression in `Description` into a UTC expression in `ScheduleExpression`.
        """
        resources = self.all_matching_resources(self.template.get("Resources", {}))

        for name, resource in resources.items():
            description = resource.get("Properties", {}).get("Description", "")
            match = aws_cron_pattern.search(description)
            if match:
                expression = "{minutes} {hours} {day_of_month} {month} {day_of_week} {year}".format(
                    **match.groupdict()
                )
                new_expression = correct_cron_expression_for_utc(expression, self.today)
                if expression != new_expression:
                    properties = resource.get("Properties")
                    properties["ScheduleExpression"] = "cron({})".format(new_expression)
                    if self.verbose:
                        print("INFO: updating {}".format(name))
                    self.dirty = True

    def main(self, tz, date, dry_run, verbose, paths):
        self.dry_run = dry_run
        self.verbose = verbose
        self.timezone = tz
        if date:
            self.today = date
        self.update(paths)


aws_cron_pattern = re.compile(
    r"cron\s*\(\s*(?P<minutes>[^\s]*)\s*(?P<hours>[^\s]*)\s*(?P<day_of_month>[^\s]*)\s*(?P<month>[^\s]*)\s*(?P<day_of_week>[^\s]*)\s*(?P<year>[^\s\)]*)\)"
)
cron_pattern = re.compile(
    r"\s*(?P<minutes>[^\s]*)\s*(?P<hours>[^\s]*)\s*(?P<day_of_month>[^\s]*)\s*(?P<month>[^\s]*)\s*(?P<day_of_week>[^\s]*)\s*(?P<year>[^\s\)]*)"
)


def correct_for_utc(hour, utcoffset):
    if utcoffset.seconds % 3600 == 0:
        return int((hour + 24 - (utcoffset.seconds / 3600)) % 24)
    else:
        raise ValueError("UTC offset is not a multiple of hours.")


def correct_cron_hours_expression_for_utc(expression, utcoffset):
    if expression == "*" or expression == "?":
        return expression

    range = re.match(r"([0-9]+)-([0-9]+)", expression)
    if range:
        from_hour = int(range.group(1))
        to_hour = int(range.group(2))
        return "{}-{}".format(
            correct_for_utc(from_hour, utcoffset), correct_for_utc(to_hour, utcoffset)
        )

    repeat = re.match(r"([0-9]+)/([0-9]+)", expression)
    if repeat:
        from_hour = int(repeat.group(1))
        repetition = int(repeat.group(2))
        return "{}/{}".format(correct_for_utc(from_hour, utcoffset), repetition)

    hours = list(filter(lambda n: re.match(r"[0-9]+", n), expression.split(",")))
    if len(hours) == len(expression.split(",")):
        hours = list(
            map(lambda h: "{}".format(correct_for_utc(int(h), utcoffset)), hours)
        )
        return ",".join(hours)
    else:
        assert False, "unsupported hour format {}".format(expression)

    return expression


def correct_cron_expression_for_utc(expression, today):
    match = cron_pattern.search(expression)
    assert match, '"{}" is not a cron expression'.format(expression)

    cron = match.groupdict()
    tomorrow_midnight = today.tzinfo.localize(
        datetime(today.year, today.month, today.day) + timedelta(days=1)
    )

    try:
        ccron = {k: ("*" if v == "?" else v) for (k, v) in cron.items()}
        expression = (
            "{minutes} {hours} {day_of_month} {month} {day_of_week} {year}".format(
                **ccron
            )
        )
        next_time = croniter(expression, tomorrow_midnight).get_next(datetime)
    except ValueError as e:
        sys.stderr.write("ERROR: {}".format(e))
        sys.exit(1)

    utcoffset = next_time.tzinfo.utcoffset(next_time)
    if utcoffset.seconds % 3600 == 0:
        cron["hours"] = correct_cron_hours_expression_for_utc(cron["hours"], utcoffset)
        expression = (
            "{minutes} {hours} {day_of_month} {month} {day_of_week} {year}".format(
                **cron
            )
        )
    else:
        sys.stderr.write(
            'WARN: UTC offset for "{}" in timezone "{}" is not a multiple of hours but {} seconds.\n'.format(
                expression, today.tzinfo, utcoffset.seconds
            )
        )

    return expression



from aws_cfn_update.replace_references import replace_references
from aws_cfn_update.cfn_updater import CfnUpdater
from io import StringIO
import textwrap


def test_preserve_comment():
    yaml = CfnUpdater().yaml

    input = "---\n# comments are great!\nResources:\nAMI: !Ref Old\n"

    template = yaml.load(input)
    result = StringIO()
    yaml.dump(template, result)
    output = result.getvalue()

    print("")
    print(input)
    print(output)
    assert output == input


def test_preserve_indent():
    yaml = CfnUpdater().yaml

    input = textwrap.dedent(
        """\
        ---
        Resources:
          Type:  # array
            - A
            - B
            - C
        """
    )

    template = yaml.load(input)
    result = StringIO()
    yaml.dump(template, result)
    output = result.getvalue()

    assert output == input


def test_preserved_tagged_array():
    yaml = CfnUpdater().yaml

    input = textwrap.dedent(
        """\
        ---
        Resources:
          Type: !Sub
            - '${B}-{A}'
            - B: Hello
              C: World
        """
    )

    template = yaml.load(input)
    result = StringIO()
    yaml.dump(template, result)
    output = result.getvalue()

    assert output == input


def test_nested_tagged_sequence_indentation():
    """
    test whether the indent of nested sequences is 4 spaces instead of 2. This is a quirk of Ruamel yaml.
    """
    input = textwrap.dedent(
        """\
    ---
    Resources:
      KongRouteCors:
        Type: Custom::KongPlugin

        Properties:
          Plugin:
            name: cors
            route:
              id: !GetAtt 'KongRoute.id'
            protocols:
              - http
              - https
            config:
              origins:
                - !Sub 'https://${ExternalDomainName}'
                - !If
                    - IsDevEnv
                    - http://localhost:3000
                    - !Ref 'AWS::NoValue'
                - !If
                    - IsDevEnv
                    - !Sub
                        - 'https://${PortalExternalDomainName}'
                        - PortalExternalDomainName: pipodeclown.nl
                    - !Ref 'AWS::NoValue'
              methods:
                - GET
                - PUT
                - POST
                - DELETE
                - PATCH
                - OPTIONS
      """
    )
    yaml = CfnUpdater().yaml

    template = yaml.load(input)
    result = StringIO()
    yaml.dump(template, result)
    output = result.getvalue()
    assert output == input



import pathlib
import tempfile

from ruamel.yaml import YAML

from aws_cfn_update.add_missing_resources import add_missing_resources
from aws_cfn_update.cfn_updater import read_template


def test_simple():
    source = {"Resources": {"AMI": {}, "EC2Instance": {"ImageId": {"Ref": "AMI"}}}}
    target = {"Resources": {"AMI": {}}}
    result = add_missing_resources(target, source)
    assert result
    assert target == source


def test_add_all():
    source = {
        "Parameters": {"VPC": {"Type": "String"}},
        "Resources": {"AMI": {}, "EC2Instance": {"ImageId": {"Ref": "AMI"}}},
        "Conditions": {"IsTrue": {"Fn::Equals": ["true", "false"]}},
        "Mappings": {"XXX": {"Type": "String"}},
    }
    target = {"Resources": {"AMI": {}}}
    result = add_missing_resources(target, source)
    assert result
    assert target == source


def test_add_no_outputs():
    source = {
        "Resources": {"AMI": {}, "EC2Instance": {"ImageId": {"Ref": "AMI"}}},
        "Outputs": {"XXX": {"Value": "bla"}},
    }
    target = {"Resources": {"AMI": {}}}
    result = add_missing_resources(target, source)
    assert target["Resources"].get("EC2Instance")
    assert not target.get("Outputs")


def test_read_template():
    source = {
        "Resources": {"AMI": {}, "EC2Instance": {"ImageId": {"Ref": "AMI"}}},
        "Outputs": {"XXX": {"Value": "bla"}},
    }
    with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
        YAML(typ="safe").dump(source, pathlib.Path(f.name))
        result = read_template(f.name)
        assert source == result



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys

import click
from ruamel.yaml.scalarstring import LiteralScalarString

from aws_cfn_update.cfn_updater import CfnUpdater


class StateMachineDefinitionUpdater(CfnUpdater):
    """
    Updates the definition of an AWS::StepFunctions::StateMachine.

    The definition is read from the file specified by --definition.
    By default, the content will be passed into the Fn::Sub function to
    allow references to parameters and resource attributes in the template.

    If you do not want substitution for your definition, specify --no-fn-sub.

    """

    def __init__(self):
        super(StateMachineDefinitionUpdater, self).__init__()
        self.resource_name = None
        self._definition = None
        self.definition_file = None
        self.template = None
        self.with_fn_sub = True
        self.verbose = False
        self.dry_run = False

    def new_value(self, definition):
        if self.with_fn_sub:
            return isinstance(definition, dict) and definition.get(
                "Fn::Sub"
            ) == self.definition, {"Fn::Sub": self.definition}
        else:
            return definition == self.definition, self.definition

    def update_template(self):
        """
        updates the Definition of Step Function StateMachine.
        """
        resources = self.template.get("Resources", {})
        for name, resource in map(
            lambda n: (n, resources[n]),
            filter(lambda n: n == self.resource_name, resources),
        ):
            if "Properties" not in resource:
                resource["Properties"] = {}

            same, new_definition = self.new_value(
                resource["Properties"].get("DefinitionString")
            )
            if not same:
                sys.stderr.write(
                    "INFO: updating definition of state machine {}\n".format(
                        self.resource_name
                    )
                )
                resource["Properties"]["DefinitionString"] = new_definition
                self.dirty = True
            else:
                if self.verbose:
                    sys.stderr.write(
                        "INFO: no changes to definition of state machine {}\n".format(
                            self.resource_name
                        )
                    )

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, definition):
        self._definition = (
            LiteralScalarString(definition)
            if not isinstance(definition, LiteralScalarString)
            else definition
        )

    def read_definition(self, filename):
        self.definition_file = filename
        with open(self.definition_file, "r") as f:
            self.definition = LiteralScalarString(f.read())

    def main(self, resource_name, definition_file, with_fn_sub, path, dry_run, verbose):
        self.dry_run = dry_run
        self.verbose = verbose
        self.resource_name = resource_name
        self.with_fn_sub = with_fn_sub
        self.read_definition(definition_file)
        self.update(path)


@click.command(
    name="state-machine-definition", help=StateMachineDefinitionUpdater.__doc__
)
@click.option(
    "--resource",
    required=True,
    help="AWS::StepFunctions::StateMachine definition to update",
)
@click.option(
    "--definition",
    required=True,
    type=click.Path(exists=True),
    help="of the state machine",
)
@click.option(
    "--fn-sub/--no-fn-sub", required=False, default=True, help="for the definition"
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def update_state_machine_definition(ctx, resource, definition, fn_sub, path):
    updater = StateMachineDefinitionUpdater()
    updater.main(
        resource, definition, fn_sub, list(path), ctx.obj["dry_run"], ctx.obj["verbose"]
    )



import pytest
from aws_cfn_update.oidc_provider_thumbprints_updater import (
    OIDCProviderThumbprintsUpdater,
)
import json


_single_oidc_provider = {
    "Resources": {
        "GitLabCom": {
            "Type": "AWS::IAM::OIDCProvider",
            "Properties": {
                "Url": "https://gitlab.com",
                "ThumbprintList": ["DEADBEEF"],
                "ClientIdList": ["https://gitlab.com"],
            },
        }
    }
}

_multiple_oidc_providers = {
    "Resources": {
        "GitLabCom": {
            "Type": "AWS::IAM::OIDCProvider",
            "Properties": {
                "Url": "https://gitlab.com",
                "ThumbprintList": ["DEADBEEF"],
                "ClientIdList": ["https://gitlab.com"],
            },
        },
        "GoogleCom": {
            "Type": "AWS::IAM::OIDCProvider",
            "Properties": {"Url": "https://accounts.google.com"},
        },
    }
}


@pytest.mark.parametrize("verbose_mode", [True, False])
def test_replace(verbose_mode):
    updater = OIDCProviderThumbprintsUpdater()
    updater.template = json.loads(json.dumps(_single_oidc_provider))

    updater.update_template()
    assert updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 1
    fingerprint = thumbprints[0]
    assert fingerprint != "DEADBEEF"

    updater.dirty = False
    updater.update_template()
    assert not updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 1
    assert fingerprint == thumbprints[0]


@pytest.mark.parametrize("verbose_mode", [True, False])
def test_append(verbose_mode):
    updater = OIDCProviderThumbprintsUpdater()
    updater.template = json.loads(json.dumps(_single_oidc_provider))
    updater.append = True

    updater.update_template()
    assert updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 2
    fingerprint = thumbprints[-1]
    assert fingerprint != "DEADBEEF"

    updater.dirty = False
    updater.update_template()
    assert not updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 2
    assert fingerprint == thumbprints[-1]


@pytest.mark.parametrize("verbose_mode", [True, False])
def test_single(verbose_mode):
    updater = OIDCProviderThumbprintsUpdater()
    updater.template = json.loads(json.dumps(_single_oidc_provider))
    updater.url = "https://accounts.google.com"

    updater.update_template()
    assert not updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 1
    fingerprint = thumbprints[0]
    assert fingerprint == "DEADBEEF"

    updater.url = "https://gitlab.com"
    updater.update_template()
    assert updater.dirty

    thumbprints = updater.template["Resources"]["GitLabCom"]["Properties"][
        "ThumbprintList"
    ]
    assert len(thumbprints) == 1
    fingerprint = thumbprints[0]
    assert fingerprint != "DEADBEEF"


@pytest.mark.parametrize("verbose_mode", [True, False])
def test_update_all(verbose_mode):
    updater = OIDCProviderThumbprintsUpdater()
    updater.template = json.loads(json.dumps(_multiple_oidc_providers))

    updater.update_template()
    assert updater.dirty



from aws_cfn_update.lambda_s3_key_updater import LambdaS3KeyUpdater as Updater
from copy import deepcopy


sample = {
    "Resources": {
        "0": {
            "Type": "AWS::Lambda::Function",
            "Properties": {
                "Code": {
                    "S3Bucket": "test-bucket",
                    "S3Key": "lambdas/cfn-listener-rule-provider-0.1.0.zip",
                },
            },
        },
        "1": {
            "Type": "AWS::Lambda::Function",
            "Properties": {
                "Code": {
                    "S3Bucket": "test-bucket",
                    "S3Key": "lambdas/iam-sudo-0.0.0.zip",
                },
            },
        },
    }
}


def test_simple_keys():
    updater = Updater()
    updater.s3_keys = [
        "lambdas/cfn-listener-rule-provider-1.0.0.zip",
    ]
    updater.template = deepcopy(sample)

    assert (
        updater.template["Resources"]["1"]["Properties"]["Code"]["S3Key"]
        == sample["Resources"]["1"]["Properties"]["Code"]["S3Key"]
    )
    updater.update_template()
    assert updater.dirty
    for i, s3_key in enumerate(updater.s3_keys):
        new_value = updater.template["Resources"][f"{i}"]["Properties"]["Code"]["S3Key"]
        assert s3_key == new_value


def test_multiple_keys():
    updater = Updater()
    updater.s3_keys = [
        "lambdas/cfn-listener-rule-provider-1.0.0.zip",
        "lambdas/iam-sudo-1.0.0.zip",
    ]
    updater.template = deepcopy(sample)

    updater.update_template()
    assert updater.dirty
    for i, s3_key in enumerate(updater.s3_keys):
        new_value = updater.template["Resources"][f"{i}"]["Properties"]["Code"]["S3Key"]
        assert s3_key == new_value


def test_non_semver_s3_key():
    updater = Updater()
    try:
        updater.s3_keys = [
            "lambdas/cfn-listener-rule-provider.zip",
        ]
        assert False, "expected invalid semver version error"
    except ValueError as error:
        assert error.args == (
            "lambdas/cfn-listener-rule-provider.zip is not a semver S3Key",
        )
        pass


def test_no_updates():
    updater = Updater()
    updater.s3_keys = [
        "lambdas/cfn-secret-provider-1.0.0.zip",
    ]
    updater.template = deepcopy(sample)
    updater.update_template()
    assert not updater.dirty



from aws_cfn_update.remove_resource import remove_resource_from_template
from aws_cfn_update.cfn_updater import CfnUpdater
from io import StringIO
from ruamel.yaml import YAML


def test_simple():
    template = {"Resources": {"AMI": {}}}
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None


def test_simple_ref():
    template = {"Resources": {"AMI": {}, "EC2Instance": {"ImageId": {"Ref": "AMI"}}}}
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("ECInstance") is None


def test_get_att():
    template = {"Resources": {"AMI": {}, "EC2Instance": {"FN::GetAtt": ["AMI", "Arn"]}}}
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_nested_ref():
    template = {
        "Resources": {"AMI": {}, "EC2Instance": {"FN::GetAtt": [{"Ref": "AMI"}, "Arn"]}}
    }
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_array_reference():
    template = {
        "Resources": {
            "AMI": {},
            "EC2Instance": {"ImageIds": [{"Ref": "AMI"}, {"Ref": "None"}]},
        }
    }
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("AMI") is None


def test_yaml():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: !Ref Old
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None


def test_yaml_simple_ref():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !Ref AMI
Outputs:
  Arn:    
     Value: !GetAtt AMI.Arn
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    resources = template.get("Resources")
    assert resources.get("AMI") is None
    assert resources.get("EC2Instance") is None
    outputs = template.get("Outputs")
    assert outputs.get("Arn") is None


def test_yaml_simple_getatt():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !GetAtt
      - AMI
      - Arn
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_yaml_nested_ref():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !GetAtt
      - !Ref AMI
      - Arn
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_yaml_ref_in_sub():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !Sub '${AMI}'
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_yaml_no_ref_in_sub():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !Sub '${!AMI}'
"""
    )

    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance")


def test_yaml_ref_in_sub_array_style():
    yaml = CfnUpdater().yaml
    template = yaml.load(
        """
Resources:
  AMI: 
    Type: Custom::AMI
  EC2Instance:
    ImageId: !Sub
      - '${AmiReference}'
      - AmiReference: !Ref AMI
"""
    )
    assert template.get("Resources", {}).get("AMI")
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_json_sub():
    template = {"Resources": {"AMI": {}, "EC2Instance": {"FN::Sub": "${AMI}"}}}
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None


def test_json_sub():
    template = {
        "Resources": {
            "AMI": {},
            "EC2Instance": {"FN::Sub": ["${AmiRef}", {"AmiRef": {"Ref": "AMI"}}]},
        }
    }
    remove_resource_from_template(template, "AMI")
    assert template.get("Resources", {}).get("AMI") is None
    assert template.get("Resources", {}).get("EC2Instance") is None



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import os
import re
import shutil
import sys
from copy import copy
from datetime import datetime
from aws_cfn_update.packer_ami_updater import main as packer_latest_ami

import click
import pytz

from aws_cfn_update.config_rule_inline_code_updater import ConfigRuleInlineCodeUpdater
from aws_cfn_update.container_image_updater import ContainerImageUpdater
from aws_cfn_update.cron_schedule_expression_updater import (
    CronScheduleExpressionUpdater,
)
from aws_cfn_update.latest_ami_updater import AMIUpdater
from aws_cfn_update.rest_api_body_updater import RestAPIBodyUpdater
from aws_cfn_update.lambda_inline_code_updater import LambdaInlineCodeUpdater
from aws_cfn_update.lambda_s3_key_updater import LambdaS3KeyUpdater
from aws_cfn_update.statemachine_updater import update_state_machine_definition
from aws_cfn_update.remove_resource import remove_resource
from aws_cfn_update.add_new_resources import add_new_resources
from aws_cfn_update.oidc_provider_thumbprints_updater import (
    update_oidc_provider_thumbprint,
)


@click.group()
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="do not change anything, just show what is going to happen",
)
@click.option("--verbose", is_flag=True, default=False, help="show more output")
@click.pass_context
def cli(ctx, dry_run, verbose):
    """Programmatically update CloudFormation templates"""
    ctx.obj = copy(ctx.params)


def validate_image(ctx, param, value):
    if all(map(lambda i: re.match(r"(?:.+/)?([^:]+)(?::.+)?", i), value)):
        return value
    else:
        raise click.BadParameter(
            '"{}" is not a valid docker image specification'.format(value)
        )


@cli.command(name="container-image", help=ContainerImageUpdater.__doc__)
@click.option(
    "--image",
    required=False,
    multiple=True,
    default=[],
    callback=validate_image,
    help="to update to",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def task_image(ctx, image, path):
    if not image:
        image = os.getenv("AWS_CFN_UPDATE_CONTAINER_IMAGES", "").split()

    if not image:
        click.echo("no container images to update")
        return

    updater = ContainerImageUpdater()
    updater.main(image, ctx.obj["dry_run"], ctx.obj["verbose"], list(path))


@cli.command(name="latest-ami", help=AMIUpdater.__doc__)
@click.option(
    "--ami-name-pattern",
    required=True,
    help="glob style pattern of the AMI image name to use",
)
@click.option(
    "--add-new-version",
    is_flag=True,
    default=False,
    help="of the AMI resource and replace all references",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def ami_image_update(ctx, ami_name_pattern, add_new_version, path):
    updater = AMIUpdater()
    updater.main(
        ami_name_pattern,
        ctx.obj["dry_run"],
        ctx.obj["verbose"],
        add_new_version,
        list(path),
    )


@cli.command(
    name="cron-schedule-expression", help=CronScheduleExpressionUpdater.__doc__
)
@click.option(
    "--timezone",
    required=False,
    help="to use to calculate the UTC time",
    default="Europe/Amsterdam",
)
@click.option(
    "--date",
    type=click.DateTime(),
    default=datetime.now(),
    help="to use as reference date",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def cron_schedule_expression(ctx, timezone, date, path):
    updater = CronScheduleExpressionUpdater()
    try:
        tz = pytz.timezone(timezone)
        updater.main(tz, date, ctx.obj["dry_run"], ctx.obj["verbose"], list(path))
    except pytz.exceptions.UnknownTimeZoneError:
        raise click.BadParameter(
            "invalid timezone specified", ctx=ctx, param="timezone"
        )


@cli.command(name="rest-api-body", help=RestAPIBodyUpdater.__doc__)
@click.option(
    "--resource", required=True, help="AWS::ApiGateway::RestApi body to update"
)
@click.option(
    "--open-api-specification",
    required=True,
    type=click.Path(exists=True),
    help="defining the interface",
)
@click.option(
    "--api-gateway-extensions",
    required=True,
    type=click.Path(exists=True),
    help="to add the the specification",
)
@click.option(
    "--add-new-version",
    is_flag=True,
    default=False,
    help="of the RestAPI resource and replace all references",
)
@click.option(
    "--keep",
    default=1,
    help="number of versions to keep, if --add-new-version is specified",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def swagger_document(
    ctx,
    resource,
    open_api_specification,
    api_gateway_extensions,
    path,
    add_new_version,
    keep,
):
    updater = RestAPIBodyUpdater()
    updater.main(
        resource,
        open_api_specification,
        api_gateway_extensions,
        list(path),
        add_new_version,
        keep,
        ctx.obj["dry_run"],
        ctx.obj["verbose"],
    )


@cli.command(name="lambda-inline-code", help=LambdaInlineCodeUpdater.__doc__)
@click.option(
    "--resource", required=True, help="name of the AWS::Lambda::Function to update"
)
@click.option(
    "--file", required=True, type=click.Path(exists=True), help="containing the source"
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def lambda_body(ctx, resource, file, path):
    updater = LambdaInlineCodeUpdater()

    with open(file, "r") as f:
        body = f.read()
    updater.main(resource, body, list(path), ctx.obj["dry_run"], ctx.obj["verbose"])


@cli.command(name="lambda-s3-key", help=LambdaS3KeyUpdater.__doc__)
@click.option(
    "--s3-key",
    required=False,
    multiple=True,
    default=[],
    help="The new S3 key in semver format",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def update_s3_key(ctx, s3_key, path):
    updater = LambdaS3KeyUpdater()
    if not s3_key:
        s3_key = os.getenv("AWS_CFN_UPDATE_LAMBDA_S3_KEYS", "").split()

    if not s3_key:
        click.echo("no Lambda s3 keys to update")
        return

    updater.main(s3_key, list(path), ctx.obj["dry_run"], ctx.obj["verbose"])


@cli.command(name="config-rule-inline-code", help=ConfigRuleInlineCodeUpdater.__doc__)
@click.option("--resource", required=True, help="name of the AWS::Config::ConfigRule")
@click.option(
    "--file", required=True, type=click.Path(exists=True), help="containing the source"
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def config_rule_body(ctx, resource, file, path):
    updater = ConfigRuleInlineCodeUpdater()

    with open(file, "r") as f:
        body = f.read()
    updater.main(resource, body, list(path), ctx.obj["dry_run"], ctx.obj["verbose"])


cli.add_command(update_state_machine_definition)
cli.add_command(remove_resource)
cli.add_command(add_new_resources)
cli.add_command(packer_latest_ami)
cli.add_command(update_oidc_provider_thumbprint)


def main():
    width = shutil.get_terminal_size().columns if sys.stdout.isatty() else 255
    cli(terminal_width=width, max_content_width=width - 5)


if __name__ == "__main__":
    main()



from aws_cfn_update.statemachine_updater import StateMachineDefinitionUpdater
import json


sample = {
    "Resources": {
        "Machine1": {
            "Type": "AWS::StepFunctions::StateMachine",
            "Properties": {"DefinitionString": ""},
        },
        "Machine2": {"Type": "AWS::StepFunctions::StateMachine"},
    }
}


def test_replace_body():
    updater = StateMachineDefinitionUpdater()
    updater.resource_name = "Machine1"
    updater.template = json.loads(json.dumps(sample))
    updater.definition = '{ "hello world": "1" }'

    updater.update_template()
    assert updater.dirty
    definition = updater.template["Resources"]["Machine1"]["Properties"].get(
        "DefinitionString"
    )
    assert isinstance(definition, dict)
    assert definition.get("Fn::Sub") == updater.definition

    updater.resource_name = "Machine2"
    updater.with_fn_sub = False
    updater.update_template()
    assert updater.dirty
    definition = updater.template["Resources"]["Machine2"]["Properties"].get(
        "DefinitionString"
    )
    assert isinstance(definition, str)
    assert definition == updater.definition



from aws_cfn_update.lambda_inline_code_updater import LambdaInlineCodeUpdater as Updater
import json


sample = {"Resources": {"Lambda": {"Type": "AWS::Lambda::Function"}}}


def test_replace_body():
    updater = Updater()
    updater.resource = "Lambda"
    updater.code = 'print("hello world!")'
    updater.template = sample.copy()

    updater.update_template()
    assert updater.dirty
    assert "Properties" in updater.template["Resources"]["Lambda"]
    assert "Code" in updater.template["Resources"]["Lambda"]["Properties"]
    assert "ZipFile" in updater.template["Resources"]["Lambda"]["Properties"]["Code"]
    assert (
        updater.template["Resources"]["Lambda"]["Properties"]["Code"]["ZipFile"]
        == 'print("hello world!")'
    )


skip_body_sample = {"Resources": {"Lambda": {"Type": "AWS::EC2::Instance"}}}


def test_skip_body():
    updater = Updater()
    updater.resource = "Lambda"
    updater.code = 'print("hello world!")'
    updater.template = skip_body_sample.copy()

    updater.update_template()
    assert not updater.dirty
    assert "Properties" not in updater.template["Resources"]["Lambda"]



import json
import os

from aws_cfn_update.rest_api_body_updater import RestAPIBodyUpdater


def test_new_resource_name():
    updater = RestAPIBodyUpdater()

    updater.resource_name = "RestAPI"
    result = updater.new_resource_name("RestAPI")
    assert result == "RestAPIv1"

    result = updater.new_resource_name(result)
    assert result == "RestAPIv2"

    updater.resource_name = "Restv2Api"
    result = updater.new_resource_name("Restv2Apiv3")
    assert result == "Restv2Apiv4"


def test_matching_names():
    updater = RestAPIBodyUpdater()
    updater.resource_name = "RestAPI"
    updater.template = {
        "Resources": {
            "RestAPIv10": {"Type": "AWS::ApiGateway::RestApi"},
            "RestApi": {"Type": "AWS::ApiGateway::RestApi"},
            "Restv2APiv1": {"Type": "AWS::ApiGateway::RestApi"},
            "RestAPIv1": {"Type": "AWS::ApiGateway::RestApi"},
            "RestAPIv0": {"Type": "AWS::ApiGateway::RestApi"},
            "RestAPIv_01": {"Type": "AWS::ApiGateway::RestApi"},
        }
    }

    result = updater.find_matching_resources()
    assert result == ["RestAPIv0", "RestAPIv1", "RestAPIv10"]

    updater.resource_name = "RestApi"
    result = updater.find_matching_resources()
    assert result == ["RestApi"]

    updater.resource_name = "NotARestApi"
    result = updater.find_matching_resources()
    assert not result


sample = {
    "Resources": {
        "Deployment": {
            "Type": "AWS::ApiGateway::Deployment",
            "Properties": {"RestApiId": {"Ref": "RestAPI"}},
        },
        "RestAPI": {
            "Type": "AWS::ApiGateway::RestApi",
            "Properties": {
                "EndpointConfiguration": {"Types": ["REGIONAL"]},
                "Body": {"swagger": "2.0"},
            },
        },
    }
}


def test_replace_body():
    updater = RestAPIBodyUpdater()
    updater.resource_name = "RestAPI"
    updater.template = json.loads(json.dumps(sample))
    updater.body = {"swagger": "2.0", "description": "a new one"}

    updater.update_template()
    assert updater.dirty
    assert (
        "description" in updater.template["Resources"]["RestAPI"]["Properties"]["Body"]
    )
    assert (
        updater.template["Resources"]["RestAPI"]["Properties"]["Body"]["description"]
        == "a new one"
    )


def test_add_new_version():
    updater = RestAPIBodyUpdater()
    updater.resource_name = "RestAPI"
    updater.template = json.loads(json.dumps(sample))
    updater.body = {"swagger": "2.0", "description": "a new one"}

    updater.add_new_version = True
    updater.update_template()
    assert updater.dirty
    assert "RestAPI" not in updater.template["Resources"]
    assert "RestAPIv1" in updater.template["Resources"]
    assert (
        "description"
        in updater.template["Resources"]["RestAPIv1"]["Properties"]["Body"]
    )
    assert (
        updater.template["Resources"]["RestAPIv1"]["Properties"]["Body"]["description"]
        == "a new one"
    )


multiple = {
    "Resources": {
        "Deployment": {
            "Type": "AWS::ApiGateway::Deployment",
            "Properties": {"RestApiId": {"Ref": "RestAPIv3"}},
        },
        "RestAPI": {
            "Type": "AWS::ApiGateway::RestApi",
            "Properties": {
                "EndpointConfiguration": {"Types": ["REGIONAL"]},
                "Body": {"swagger": "2.0"},
            },
        },
        "RestAPIv2": {
            "Type": "AWS::ApiGateway::RestApi",
            "Properties": {
                "EndpointConfiguration": {"Types": ["REGIONAL"]},
                "Body": {"swagger": "2.0"},
            },
        },
        "RestAPIv3": {
            "Type": "AWS::ApiGateway::RestApi",
            "Properties": {
                "EndpointConfiguration": {"Types": ["GLOBAL"]},
                "Body": {"swagger": "2.0"},
            },
        },
    }
}


def test_add_new_version_keep_two():
    updater = RestAPIBodyUpdater()
    updater.resource_name = "RestAPI"
    updater.template = json.loads(json.dumps(multiple))
    updater.body = {"swagger": "2.0", "description": "a new one"}

    updater.add_new_version = True
    updater.keep = 2
    updater.update_template()
    assert updater.dirty
    assert "RestAPI" not in updater.template["Resources"]
    assert "RestAPIv2" not in updater.template["Resources"]
    assert "RestAPIv3" in updater.template["Resources"]
    assert "RestAPIv4" in updater.template["Resources"]
    assert (
        "GLOBAL"
        == updater.template["Resources"]["RestAPIv4"]["Properties"][
            "EndpointConfiguration"
        ]["Types"][0]
    )
    assert (
        "description"
        in updater.template["Resources"]["RestAPIv4"]["Properties"]["Body"]
    )
    assert (
        updater.template["Resources"]["RestAPIv4"]["Properties"]["Body"]["description"]
        == "a new one"
    )

    assert (
        updater.template["Resources"]["Deployment"]["Properties"]["RestApiId"]["Ref"]
        == "RestAPIv4"
    )


def test_load_and_merge():
    test_directory = os.path.dirname(os.path.abspath(__file__))
    updater = RestAPIBodyUpdater()
    updater.resource_name = "RestAPI"
    updater.template = json.loads(json.dumps(sample))
    updater.api_gateway_extensions = f"{test_directory}/aws-extensions.yaml"
    updater.open_api_specification = f"{test_directory}/api-specification.yaml"
    updater.body = updater.load_and_merge_swagger_body()
    updater.update_template()

    assert updater.dirty
    assert (
        updater.template["Resources"]["RestAPI"]["Properties"]["Body"] == updater.body
    )



import textwrap

from aws_cfn_update.container_image_updater import ContainerImageUpdater as Updater
from copy import deepcopy


sample = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Resources": {
        "AMI": {
            "Type": "Custom::AMI",
            "Properties": {
                "Filters": {"name": "amzn-ami-2017.09.a-amazon-ecs-optimized"}
            },
        },
        "TaskDefinition": {
            "Type": "AWS::ECS::TaskDefinition",
            "Properties": {
                "Family": "testme",
                "NetworkMode": "bridge",
                "ContainerDefinitions": [
                    {
                        "Name": "paas-monitor",
                        "Image": "mvanholsteijn/paas-monitor:latest",
                    },
                    {"Name": "sidecar", "Image": "cloud_sql_proxy:latest"},
                ],
            },
        },
    },
}


def test_single_update():
    new_images = ["mvanholsteijn/paas-monitor:0.6.0"]
    updater = Updater()
    updater.filename = "test.json"
    updater.images = new_images
    updater.template = deepcopy(sample)

    updater.update_template()
    assert updater.dirty
    for image in new_images:
        new_value = updater.template["Resources"]["TaskDefinition"]["Properties"][
            "ContainerDefinitions"
        ][0]["Image"]
        assert image == new_value


def test_multiple_update():
    new_images = ["mvanholsteijn/paas-monitor:0.6.0", "cloud_sql_proxy:3.3.0"]
    updater = Updater()
    updater.filename = "test.json"
    updater.images = new_images
    updater.template = deepcopy(sample)

    updater.update_template()
    assert updater.dirty
    for i, image in enumerate(new_images):
        new_value = updater.template["Resources"]["TaskDefinition"]["Properties"][
            "ContainerDefinitions"
        ][i]["Image"]
        assert image == new_value


def test_no_updates():
    updater = Updater()
    updater.images = [
        "lambdas/cfn-secret-provider:1.0.0",
    ]
    updater.template = deepcopy(sample)
    updater.update_template()
    assert not updater.dirty


def test_invalid_container_image():
    updater = Updater()
    try:
        updater.images = [
            "alpine",
        ]
    except ValueError as error:
        assert error.args == ("alpine is an invalid image name",)


def test_container_image_reference_not_found():
    template = textwrap.dedent(
        """
        AWSTemplateFormatVersion: '2010-09-09'
        Parameters:
          PaasMonitorImage: String
        
        Resources:
          AMI:
            Type: Custom::AMI
            Properties:
              Filters:
                name: amzn-ami-2017.09.a-amazon-ecs-optimized
          TaskDefinition:
            Type: AWS::ECS::TaskDefinition
            Properties:
              Family: testme
              NetworkMode: bridge
              ContainerDefinitions:
                - Name: paas-monitor
                  Image: !Ref PaasMonitorImage

    """
    )
    new_images = ["mvanholsteijn/paas-monitor:0.6.0"]

    updater = Updater()
    updater.filename = "test.yaml"
    updater.images = new_images
    updater.template = updater.yaml.load(template)

    updater.update_template()
    assert not updater.dirty



import pytz
from datetime import datetime, tzinfo
from aws_cfn_update.cron_schedule_expression_updater import (
    aws_cron_pattern,
    CronScheduleExpressionUpdater as Updater,
)
from aws_cfn_update.cron_schedule_expression_updater import (
    correct_for_utc,
    correct_cron_hours_expression_for_utc,
    correct_cron_expression_for_utc,
)


def test_cron_pattern_match():
    description = """
    run every hour cron(0 4 * * * *)
    """
    assert aws_cron_pattern.search(description)


def test_hour_correct_for_utc():
    summer = pytz.timezone("Europe/Berlin").localize(datetime(2018, 8, 1))

    assert 22 == correct_for_utc(0, summer.utcoffset())
    assert 23 == correct_for_utc(1, summer.utcoffset())
    assert 0 == correct_for_utc(2, summer.utcoffset())
    assert 10 == correct_for_utc(12, summer.utcoffset())
    assert 21 == correct_for_utc(23, summer.utcoffset())

    winter = pytz.timezone("Europe/Berlin").localize(datetime(2018, 12, 1))

    assert 23 == correct_for_utc(0, winter.utcoffset())
    assert 0 == correct_for_utc(1, winter.utcoffset())
    assert 1 == correct_for_utc(2, winter.utcoffset())
    assert 11 == correct_for_utc(12, winter.utcoffset())
    assert 22 == correct_for_utc(23, winter.utcoffset())


def test_correct_cron_hours_expression_for_utc():
    summer = pytz.timezone("Europe/Berlin").localize(datetime(2018, 8, 1))

    assert "*" == correct_cron_hours_expression_for_utc("*", summer.utcoffset())
    assert "?" == correct_cron_hours_expression_for_utc("?", summer.utcoffset())
    assert "20" == correct_cron_hours_expression_for_utc("22", summer.utcoffset())
    assert "20/3" == correct_cron_hours_expression_for_utc("22/3", summer.utcoffset())
    assert "7-15" == correct_cron_hours_expression_for_utc("9-17", summer.utcoffset())
    assert "23,1,8,11" == correct_cron_hours_expression_for_utc(
        "1,3,10,13", summer.utcoffset()
    )

    winter = pytz.timezone("Europe/Berlin").localize(datetime(2018, 12, 1))

    assert "*" == correct_cron_hours_expression_for_utc("*", winter.utcoffset())
    assert "?" == correct_cron_hours_expression_for_utc("?", winter.utcoffset())
    assert "21" == correct_cron_hours_expression_for_utc("22", winter.utcoffset())
    assert "21/3" == correct_cron_hours_expression_for_utc("22/3", winter.utcoffset())
    assert "8-16" == correct_cron_hours_expression_for_utc("9-17", winter.utcoffset())
    assert "0,2,9,12" == correct_cron_hours_expression_for_utc(
        "1,3,10,13", winter.utcoffset()
    )


def test_correct_cron_expression_for_utc():
    timezone = pytz.timezone("Europe/Berlin")
    summer = timezone.localize(datetime(2018, 8, 1))

    assert "* * * * * *" == correct_cron_expression_for_utc("* * * * * *", summer)
    assert "* 2 * * * *" == correct_cron_expression_for_utc("* 4 * * * *", summer)
    assert "* 7-15 * * * *" == correct_cron_expression_for_utc("* 9-17 * * * *", summer)
    assert "* 23/3 * * * *" == correct_cron_expression_for_utc("* 1/3 * * * *", summer)
    assert "* 23/3 * * ? *" == correct_cron_expression_for_utc("* 1/3 * * ? *", summer)
    assert "* 23/3 ? * ? *" == correct_cron_expression_for_utc("* 1/3 ? * ? *", summer)
    assert "* 23/3 ? * SUN *" == correct_cron_expression_for_utc(
        "* 1/3 ? * SUN *", summer
    )


def test_template_update():
    template = {
        "Resources": {
            "Schedule": {
                "Type": "AWS::Events::Rule",
                "Properties": {
                    "Description": "Run every hour - cron(0 1/3 * * * *)",
                    "ScheduleExpression": "cron(Minutes Hours Day-of-month Month Day-of-week Year)",
                },
            }
        }
    }
    updater = Updater()
    updater.template = template
    updater.timezone = pytz.timezone("Europe/Amsterdam")
    updater.today = datetime(2018, 8, 1)
    updater.update_template()
    assert updater.dirty
    properties = template.get("Resources").get("Schedule").get("Properties")
    assert properties["ScheduleExpression"] == "cron(0 23/3 * * * *)"


def test_template_update_incorrect_cron():
    template = {
        "Resources": {
            "Schedule": {
                "Type": "AWS::Events::Rule",
                "Properties": {
                    "Description": "Run every hour - cron(1/3 * * * *)",
                    "ScheduleExpression": "cron(Minutes Hours Day-of-month Month Day-of-week Year)",
                },
            }
        }
    }
    updater = Updater()
    updater.template = template
    updater.timezone = pytz.timezone("Europe/Amsterdam")
    updater.today = datetime(2018, 8, 1)
    updater.update_template()
    assert not updater.dirty
    properties = template.get("Resources").get("Schedule").get("Properties")
    assert (
        properties["ScheduleExpression"]
        == "cron(Minutes Hours Day-of-month Month Day-of-week Year)"
    )



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from .cfn_updater import CfnUpdater
from ruamel.yaml.scalarstring import PreservedScalarString


class LambdaInlineCodeUpdater(CfnUpdater):
    """
        Updates the inline code of an AWS::Lambda::Function resource.

    \b
      ELBListenerRuleProvider:
        Type: AWS::Lambda::Function

    \b
      ELBListenerRuleProvider:
        Type: AWS::Lambda::Function
        Properties:
          Code:
            ZipFile:
              import boto3
              import cfnresponse
              ELB = boto3.client('elbv2')
              ...
    """

    def __init__(self):
        super(LambdaInlineCodeUpdater, self).__init__()
        self._image = []

    def update_template(self):
        """
        updates the Code property of a AWS::Lambda::Function resource of name `self.resource` to `self.code`
        """
        resource = self.template.get("Resources", {}).get(self.resource, None)
        if resource and resource["Type"] == "AWS::Lambda::Function":
            code = resource.get("Properties", {}).get("Code", {})
            old_code = code["ZipFile"] if "ZipFile" in code else None
            if old_code != self.code:
                sys.stderr.write(
                    "INFO: updating inline code of lambda {} in {}\n".format(
                        self.resource, self.filename
                    )
                )
                if "Properties" not in resource:
                    resource["Properties"] = {}
                if "Code" not in resource["Properties"]:
                    resource["Properties"]["Code"] = {}
                resource["Properties"]["Code"] = {
                    "ZipFile": PreservedScalarString(self.code)
                }
                self.dirty = True
        elif resource:
            sys.stderr.write(
                "WARN: resource {} in {} is not of type AWS::Lambda::Function\n".format(
                    self.resource, self.filename
                )
            )

    def main(self, resource, code, paths, dry_run, verbose):
        self.resource = resource
        self.code = code
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from .cfn_updater import CfnUpdater
from ruamel.yaml.scalarstring import PreservedScalarString


class ConfigRuleInlineCodeUpdater(CfnUpdater):
    """
        Updates the inline code of an AWS::Config::ConfigRule resource.

    \b
      ConfigRule:
        Type: AWS::Config::ConfigRule
        Properties:
          Source:
            Owner: CUSTOM_POLICY

    \b
      ConfigRule:
        Type: AWS::Config::ConfigRule
        Properties:
          Source:
            Owner: CUSTOM_POLICY
            CustomPolicyDetails:
              EnableDebugLogDelivery: true
              PolicyRuntime: guard-2.x.x
              PolicyText: |-
                rule name when resourceType == "AWS::S3::Bucket" {
                    ...
                }
              ...
    """

    def __init__(self):
        super(ConfigRuleInlineCodeUpdater, self).__init__()
        self._image = []
        self.resource_name = None
        self.code = None

    def update_template(self):
        """
        updates the Code property of a AWS::Config::ConfigRule resource of name `self.resource` to `self.code`
        """
        resource = self.template.get("Resources", {}).get(self.resource_name, None)
        properties = resource.get("Properties", {})
        if (
            resource
            and resource["Type"] == "AWS::Config::ConfigRule"
            and properties.get("Source", {}).get("Owner") == "CUSTOM_POLICY"
        ):
            source = properties.get("Source", {})
            details = source.get("CustomPolicyDetails", {})
            old_code = details.get("PolicyText", None)

            if old_code != self.code:
                sys.stderr.write(
                    "INFO: updating policy text of config rule {} in {}\n".format(
                        self.resource_name, self.filename
                    )
                )
                if "Properties" not in resource:
                    resource["Properties"] = {}
                if "Source" not in resource["Properties"]:
                    resource["Properties"]["Source"] = {}
                if "CustomPolicyDetails" not in resource["Properties"]["Source"]:
                    resource["Properties"]["Source"]["CustomPolicyDetails"] = {}

                resource["Properties"]["Source"]["CustomPolicyDetails"][
                    "PolicyText"
                ] = PreservedScalarString(self.code)
                self.dirty = True
        elif resource:
            sys.stderr.write(
                "WARN: resource {} in {} is not a CUSTOM_POLICY of type AWS::Config::ConfigRule\n".format(
                    self.resource_name, self.filename
                )
            )

    def main(self, resource, code, paths, dry_run, verbose):
        self.resource_name = resource
        self.code = code
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



from aws_cfn_update.replace_references import replace_references
from aws_cfn_update.cfn_updater import CfnUpdater
from io import StringIO


def test_simple():
    template = {"Ref": "Old"}
    assert replace_references(template, "Old", "New")
    assert template["Ref"] == "New"

    template = {"DependsOn": [{"Ref": "Old"}]}
    assert replace_references(template, "Old", "New")
    assert template["DependsOn"][0]["Ref"] == "New"

    template = "Old"
    assert not replace_references(template, "Old", "New")
    assert template == "Old"

    template = {"DependsOn": [{"Ref": "Old"}]}
    assert not replace_references(template, "Oud", "New")
    assert template["DependsOn"][0]["Ref"] == "Old"


def test_yaml():
    yaml = CfnUpdater().yaml
    template = yaml.load("---\nAMI: !Ref Old")
    assert replace_references(template, "Old", "New")

    result = StringIO()
    yaml.dump(template, result)
    assert result.getvalue() == "---\nAMI: !Ref New\n"
